--=============================================================================--
--===========================================================================--
library ieee;
   use ieee.std_logic_1164.all;
   use IEEE.STD_LOGIC_ARITH.ALL;
   use IEEE.STD_LOGIC_UNSIGNED.ALL;
   use ieee.numeric_std.all;
	
use work.pkg_vectrex.all;

entity vectrex is
  port(
		cpu_areset   : in    std_logic;
		cpu_ahold    : in    std_logic;
		cpu_anmi     : in    std_logic;
		cpu_clk		: in    std_logic; --12 MHz
		analog_sound_out    : out std_logic_vector(7 downto 0);
		analog_x_out    : out std_logic_vector(9 downto 0);
		analog_y_out    : out std_logic_vector(9 downto 0);
		analog_z_out    : out std_logic_vector(7 downto 0);
		p1_x				 : in std_logic_vector(7 downto 0);
		p1_y				 : in std_logic_vector(7 downto 0);
		p1_buttons		 : in std_logic_vector(3 downto 0);
		p2_x				 : in std_logic_vector(7 downto 0);
		p2_y				 : in std_logic_vector(7 downto 0);
		p2_buttons		 : in std_logic_vector(3 downto 0)
	);
end vectrex;

-------------------------------------------------------------------------------
-- Architecture for Vectrex
-------------------------------------------------------------------------------
architecture behaviour of vectrex is
  signal sys_clk       	: std_logic; 
  signal resetsr 		: std_logic_vector(1 downto 0);
  signal ares			: std_logic;
  signal cnt1m5			: integer range 0 to 7; 
  signal clk1m5			: std_logic;
  signal cnt6m			: integer range 0 to 1; 
  signal clk6m			: std_logic;
  signal cnt750k		: integer range 0 to 7; 

  
  -- CPU Interface signals
  signal cpu_rst       : std_logic;
  signal cpu_rw        : std_logic;
  signal cpu_vma       : std_logic;
  signal cpu_firq      : std_logic;
  signal cpu_addr      : std_logic_vector(15 downto 0);
  signal cpu_data_in   : std_logic_vector(7 downto 0);
  signal cpu_data_out  : std_logic_vector(7 downto 0);
  signal cpu_pc		   : std_logic_vector(15 downto 0);

  -- BOOT ROM
  signal rom_cs        : Std_logic;
  signal rom_data_out  : Std_Logic_Vector(7 downto 0);

  -- IO / VIA
  signal io_cs       : std_logic;
  signal io_data_out : std_logic_vector(7 downto 0);
  signal io_data_oe  : std_logic;
  signal via_wen	 : std_logic;
  signal via_ren     : std_logic;
  signal via_pa_out  : std_logic_vector(7 downto 0);
  signal via_pa_oe	 : std_logic_vector(7 downto 0);
  signal via_pa_in   : std_logic_vector(7 downto 0);
  signal via_pb_out  : std_logic_vector(7 downto 0);
  signal via_pb_oe	 : std_logic_vector(7 downto 0);
  signal via_pb_in	 : std_logic_vector(7 downto 0);
--signal via_irq_out : std_logic; -- ?????????????
  signal via_ca1_in  : std_logic; -- to IO7 AY8912
  signal via_ca2_out : std_logic; -- alg n_zero
--signal via_ca2_oe  : std_logic; -- alg n_zero
  signal via_ca2_in  : std_logic; -- alg n_zero
  signal via_cb1_out : std_logic; -- nc
--signal via_cb1_oe  : std_logic; -- nc
  signal via_cb1_in  : std_logic; -- nc
  signal via_cb2_out : std_logic; -- alg n_blank
--signal via_cb2_oe  : std_logic; -- alg n_blank
  signal via_cb2_in  : std_logic; -- alg n_blank
  
  -- Sound / AY
  signal ay_bc1		 : std_logic; -- to AY bc1
  signal ay_bdir     : std_logic; -- to AY bdir
  signal ay_pa_out	 : std_logic_vector(7 downto 0);
  signal ay_pa_in  	 : std_logic_vector(7 downto 0);
  signal ay_data_out : std_logic_vector(7 downto 0);
  signal ay_data_in  : std_logic_vector(7 downto 0);
--signal ay_data_oe  : std_logic;
  signal analog_sound    : std_logic_vector(7 downto 0);
  
  -- CART
  signal cart_cs       : std_logic;
  signal cart_data_out : std_logic_vector(7 downto 0);

  -- RAM
  signal sram_cs       : std_logic;
  signal sram_data_out : std_logic_vector(7 downto 0);
  signal sram_ce       : std_logic;
  signal sram_ub       : std_logic;
  signal sram_lb       : std_logic;
  signal sram_we       : std_logic;
  signal sram_oe       : std_logic;
  
  --Analog
  signal analog_vec    : std_logic_vector(7 downto 0);
  signal analog_x      : std_logic_vector(7 downto 0);
  signal analog_y      : std_logic_vector(7 downto 0);
  signal analog_y_temp : std_logic_vector(7 downto 0);
  signal analog_x_int  : std_logic_vector(9 downto 0);
  signal analog_y_int  : std_logic_vector(9 downto 0);
  signal analog_z      : std_logic_vector(7 downto 0);
  signal analog_z_sh : std_logic_vector(7 downto 0);
  signal analog_offs   : std_logic_vector(7 downto 0);
  signal ramp_int	   : std_logic;
  signal zero_int	   : std_logic;
  signal comparator_in : std_logic_vector(7 downto 0);
  signal compare_out   : std_logic;
--signal cs_x          : std_logic;
--signal cs_y          : std_logic;
--signal cs_x_int      : std_logic;
--signal cs_y_int      : std_logic;
--signal cs_z          : std_logic;


-----------------------------------------------------------------
--
-- CPU09 CPU core
--
-----------------------------------------------------------------

component cpu09
  port (    
	 clk      :	in  std_logic;
     rst      : in  std_logic;
     vma      :	out std_logic;
     addr     : out std_logic_vector(15 downto 0);
     --address:	 out std_logic_vector(15 downto 0);
     rw       :	out std_logic;
     data_in  : in	 std_logic_vector(7 downto 0);
	 data_out : out std_logic_vector(7 downto 0);
	 halt     : in  std_logic;
	 hold     : in  std_logic;
	 irq      : in  std_logic;
	 nmi      : in  std_logic;
	 firq     : in  std_logic
	 --dpc      : out std_logic_vector (15 downto 0)
  );
end component;

----------------------------------------
--
-- 8KByte Block RAM as Main ROM
--
----------------------------------------
component mon_rom
  Port (
    clk      : in  std_logic;
    rst      : in  std_logic;
    cs       : in  std_logic;
    rw       : in  std_logic;
    addr     : in  std_logic_vector (12 downto 0);
    data_in  : in  std_logic_vector (7 downto 0);
    data_out : out std_logic_vector (7 downto 0)
    );
end component;

component cart_rom
  Port (
    clk      : in  std_logic;
    rst      : in  std_logic;
    cs       : in  std_logic;
    rw       : in  std_logic;
    addr     : in  std_logic_vector (12 downto 0);
    data_in  : in  std_logic_vector (7 downto 0);
    data_out : out std_logic_vector (7 downto 0)
    );
end component;

component ram_2k
 Port (
    clk      : in  std_logic;
    rst      : in  std_logic;
    cs       : in  std_logic;
    rw       : in  std_logic;
    addr     : in  std_logic_vector (9 downto 0);
    data_in  : in  std_logic_vector (7 downto 0);
    data_out : out std_logic_vector (7 downto 0)
    );
end component;

component via6522 is
port (
    clock       : in  std_logic;
    clock_en    : in  std_logic; -- for counters and stuff
    reset       : in  std_logic;
    
    addr        : in  std_logic_vector(3 downto 0);
    wen         : in  std_logic;
    ren         : in  std_logic;
    data_in     : in  std_logic_vector(7 downto 0);
    data_out    : out std_logic_vector(7 downto 0);

    -- pio --
    port_a_o    : out std_logic_vector(7 downto 0);
    port_a_t    : out std_logic_vector(7 downto 0);
    port_a_i    : in  std_logic_vector(7 downto 0);
    
    port_b_o    : out std_logic_vector(7 downto 0);
    port_b_t    : out std_logic_vector(7 downto 0);
    port_b_i    : in  std_logic_vector(7 downto 0);

    -- handshake pins
    ca1_i       : in  std_logic;

    ca2_o       : out std_logic;
    ca2_i       : in  std_logic;
    ca2_t       : out std_logic;
    
    cb1_o       : out std_logic;
    cb1_i       : in  std_logic;
    cb1_t       : out std_logic;
    
    cb2_o       : out std_logic;
    cb2_i       : in  std_logic;
    cb2_t       : out std_logic;

    irq         : out std_logic
      );
end component;


component signed_integ is
      Port ( 
			  clk_i              :  in  std_logic;
			  rst_i              :  in  std_logic;
			  ramp_i			 :  in  std_logic;
			  zero_i			 :  in  std_logic;
			  data_i          	 :  in  std_logic_vector(7 downto 0);
			  offset_i			 :  in  std_logic_vector(7 downto 0);
			  data_o        	 :  out std_logic_vector(9 downto 0)
            );
end component;

component YM2149 is
	Port (
		  -- data bus
		  I_DA                : in  std_logic_vector(7 downto 0);
		  O_DA                : out std_logic_vector(7 downto 0);
		  O_DA_OE_L           : out std_logic;
		  -- control
		  I_A9_L              : in  std_logic;
		  I_A8                : in  std_logic;
		  I_BDIR              : in  std_logic;
		  I_BC2               : in  std_logic;
		  I_BC1               : in  std_logic;
		  I_SEL_L             : in  std_logic;

		  O_AUDIO             : out std_logic_vector(7 downto 0);
		  -- port a
		  I_IOA               : in  std_logic_vector(7 downto 0);
		  O_IOA               : out std_logic_vector(7 downto 0);
		  O_IOA_OE_L          : out std_logic;
		  -- port b
		  I_IOB               : in  std_logic_vector(7 downto 0);
		  O_IOB               : out std_logic_vector(7 downto 0);
		  O_IOB_OE_L          : out std_logic;
		  --
		  ENA                 : in  std_logic; -- clock enable
		  RESET_L             : in  std_logic; -- E clock , 1.5MHz
		  CLK                 : in  std_logic  -- 6 Mhz
		);
end component;


begin
  -----------------------------------------------------------------------------
  -- Instantiation of internal components
  -----------------------------------------------------------------------------

my_cpu : cpu09  port map (    
	 clk	   => clk1m5,
	 rst       => cpu_rst,
     vma       => cpu_vma,
     addr      => cpu_addr(15 downto 0),
     rw	       => cpu_rw,
     data_in   => cpu_data_in,
	 data_out  => cpu_data_out,
	 halt      => '0',
	 hold      => '0',
	 irq       => '0', 
	 nmi       => '0', 
	 firq      => '0'
    );

my_rom : mon_rom port map (
    clk       => cpu_clk,
    rst       => cpu_rst,
	cs        => rom_cs,
	rw        => '1',
    addr      => cpu_addr(12 downto 0),
    data_in   => cpu_data_out,
    data_out  => rom_data_out
    );
    
my_cart : cart_rom port map (
    clk       => cpu_clk,
    rst       => cpu_rst,
	cs        => cart_cs,
	rw        => '1',
    addr      => cpu_addr(12 downto 0),
    data_in   => cpu_data_out,
    data_out  => cart_data_out
    );
    
my_ram : ram_2k port map (
	clk       => cpu_clk,
    rst       => cpu_rst,
	cs        => sram_cs,
	rw        => cpu_rw,
    addr      => cpu_addr(9 downto 0),
    data_in   => cpu_data_out,
    data_out  => sram_data_out
	);

my_via : via6522 port map (
			clock       => clk6m,
        clock_en    => clk1m5,
        reset       => cpu_rst,
                                
        addr        => cpu_addr(3 downto 0),
        wen         => via_wen,
        ren         => via_ren,
        data_in     => cpu_data_out,
        data_out    => io_data_out,
                                
        -- pio --   
        port_a_o    => via_pa_out,
        port_a_t    => via_pa_oe,
        port_a_i    => via_pa_in,
                                
        port_b_o    => via_pb_out,
        port_b_t    => via_pb_oe,
        port_b_i    => via_pb_in,
    
        -- handshake pins
        ca1_i       => ay_pa_out(7),--via_ca1_in,
                            
        ca2_o       => via_ca2_out,
        ca2_i       => via_ca2_in,
--      ca2_t       => via_ca2_oe,
                            
        cb1_o       => via_cb1_out,
        cb1_i       => via_cb1_in,
--      cb1_t       => via_cb1_oe,
                            
        cb2_o       => via_cb2_out,
        cb2_i       => via_cb2_in
--      cb2_t       => via_cb2_oe,
                            
--      irq         => via_irq_out
    );


my_ay : YM2149
	port map (
	-- data bus
    I_DA                => ay_data_in,
    O_DA                => ay_data_out,
 --   O_DA_OE_L           => ay_data_oe,
    -- control
    I_A9_L              => '0',
    I_A8                => '1',
    I_BDIR              => ay_bdir,
    I_BC2               => '1',
    I_BC1               => ay_bc1,
    I_SEL_L             => '1',

    O_AUDIO             => analog_sound,
    -- port a
    I_IOA               => ay_pa_in,	--joysticks
    O_IOA               => ay_pa_out,
    O_IOA_OE_L          => open,		--ioa_oe_l,
    -- port b
    I_IOB               => "00000000", 	--portb nc at AY38912
    O_IOB               => open, 		--portb nc at AY38912
    O_IOB_OE_L          => open, 		--portb nc at AY38912

    ENA                 => '1',
    RESET_L             => not cpu_rst,
    CLK                 => clk1m5
    );


x_int : signed_integ
	port map (
			clk_i      		=> cpu_clk,
			rst_i      		=> cpu_rst,
			ramp_i	   		=> not ramp_int,
			zero_i	   		=> not zero_int,
			data_i			=> analog_x(7 downto 0),
			offset_i		=> analog_offs(7 downto 0),
			data_o			=> analog_x_out
			);
			
y_int : signed_integ
	port map (
			clk_i      		=> cpu_clk,
			rst_i      		=> cpu_rst,
			ramp_i	   		=> not ramp_int,
			zero_i	   		=> not zero_int,
			data_i			=> analog_y(7 downto 0),
			offset_i		=> analog_offs(7 downto 0),
			data_o			=> analog_y_out
			);
        
----------------------------------------------------------------------
--
-- Process to decode memory map
--
----------------------------------------------------------------------

mem_decode: process(cpu_clk)
begin
	if rising_edge (cpu_clk) then

		  cpu_rst    <= resetsr(1); -- CPU reset is active high	  
		  ares		 <= cpu_areset;
		
		if cpu_rst = '1' then
		  cpu_data_in <= (others=>'1');
		  rom_cs      <= '0';
		  io_cs       <= '0';
		  sram_cs     <= '0';
		  cart_cs	  <= '0';      
		else
		 

	-- Memory Decoder
		if cpu_addr( 15 ) = '0' then 				--CART 0x0000-0x7FFF
			cart_cs     <= cpu_vma;
--			if sw(1) = '1' then
				cpu_data_in <= "00000000";
				cart_cs <= '0';
--			else
--				cart_cs     <= cpu_vma;
--				cpu_data_in <= cart_data_out;
--			end if;
		else
			--cpu_data_in <= cpu_data_in;
			cart_cs <= '0';
		end if;  

		if cpu_addr( 15 downto 11 ) = "11001" then 	--RAM 0xC800-0xCFFF
			cpu_data_in <= sram_data_out;
			sram_cs     <= cpu_vma;
		else
			sram_cs <= '0';
		end if;
		
		if cpu_addr( 15 downto 12 ) = "1101" then 	--IO 0xD000-0xDFFF	    
			cpu_data_in <= io_data_out;	--VHDL VIA
			--cpu_data_in <= via_data_out;	--Real VIA
			io_cs     <= cpu_vma;
		else
			--cpu_data_in <= cpu_data_in;
			io_cs <= '0';	
		end if;

		if cpu_addr( 15 downto 13 ) = "111" then 	--ROM 0xE000-0xFFFF
			cpu_data_in <= rom_data_out;
			rom_cs      <= cpu_vma;
		else
			--cpu_data_in <= cpu_data_in;
			rom_cs <= '0';
		end if;
	end if;		
  end if;      
end process;



update_alg : process( cpu_clk )
begin
	if rising_edge (cpu_clk) then
		if cpu_rst = '1' then		
			ay_bc1		   			<= '0';
			ay_bdir        			<= '0';
			via_pa_in				<= "00000000";
			via_pb_in(5)			<= '1';
			via_pb_in(6)			<= '1';		
		else			
			via_pb_in	<= via_pb_out;		--portb via	
			via_cb1_in	<= via_cb1_out;		--via cb1
			via_pb_in(5) <= compare_out;
			ay_bc1		<= via_pb_out(3);	--AY cs
			ay_bdir     <= via_pb_out(4);	--AY direction
			via_ca2_in	<= via_ca2_out; 	--nZero			
			via_cb2_in	<= via_cb2_out;		--nBlank, Gruen			
			analog_x 	<= via_pa_out;		--analog_vec;

			zero_int	<= via_ca2_out;
			ramp_int	<= via_pb_out(7);

			if via_cb2_out = '0' then		--blanking
				analog_z <= "00000000";
			else
				analog_z <= analog_z_sh;
			end if;

			-- mux comparator in
			if (via_pb_out(2 downto 1)="00") then
				comparator_in<=p1_x;
			elsif	(via_pb_out(2 downto 1)="01") then
				comparator_in<=p1_y;
			elsif	(via_pb_out(2 downto 1)="10") then
				comparator_in<=p2_x;
			else -- "11"
				comparator_in<=p2_y;
			end if;
			
			if via_pb_out(0) = '0'  then 	-- mux enable
			
				--demux y channel 			
				if 	(via_pb_out(1) = '0') and (via_pb_out(2) = '0') then
					analog_y 	<= analog_vec;
				end if;			
				
				--demux z channel
				if 	(via_pb_out(1) = '0') and (via_pb_out(2) = '1') then
					analog_z_sh 	<= analog_vec;
				end if;
				
				--demux offset channel
				if 	(via_pb_out(1) = '1') and (via_pb_out(2) = '0') then
					analog_offs <= analog_vec;
				end if;
				
			end if;
			
			if via_pa_oe = "00000000" then		--Read VIA PortA	
				
					if via_pb_out(3) = '1' and via_pb_out(4) = '0' then
						via_pa_in		<= ay_data_out;
					else
						via_pa_in	    <=  analog_vec;
					end if;						
			else								--Write VIA PortA				
					via_pa_in		<= via_pa_out;	
					ay_data_in		<= via_pa_out;
					analog_vec		<= via_pa_out;				
			end if;			
		end if;
	end if;	
end process;

update_joys: process(cpu_clk) --Synchronous because we handle user input here.
begin
   if rising_edge(cpu_clk) and clk6m='1' then
		if comparator_in>analog_vec then 
			compare_out<='0';
		else
			compare_out<='1';
		end if;
	end if;
end process;


reset : process begin 
   wait until rising_edge(clk1m5); -- and pll_locked = '1';
   resetsr <= resetsr(0) & ares;
end process;

vex_clk : process (cpu_clk,clk6m,clk1m5)
begin
	if rising_edge(cpu_clk) then
		if cnt6m = 2-1 then
		   clk6m  <= '1';
		   cnt6m  <= 0;
		else
		   clk6m  <= '0';
		   cnt6m  <= cnt6m +1;
		end if;
	 end if; 
   
	if rising_edge(clk6m) then       
		if cnt1m5  = 4-1 then
		   clk1m5  <= '1';
		   cnt1m5  <= 0;
		else
		   clk1m5  <= '0';
		   cnt1m5  <= cnt1m5 +1;
		end if;
	 end if;   
end process;


analog_z_out<=analog_z;
analog_sound_out<=analog_sound;
via_wen<=not cpu_rw and io_cs and cpu_addr(12);
via_ren<=cpu_rw and io_cs and cpu_addr(12);
ay_pa_in<=p2_buttons&p1_buttons;

end behaviour; --===================== End of architecture =======================--

